from soar.widgets import *

import io
reload(io)
from io import *

import GridWorld
reload(GridWorld)
from GridWorld import *

import Planner
reload(Planner)
from Planner import *

import XYThDriver
reload(XYThDriver)
from XYThDriver import *

import TSMFull
reload(TSMFull)
from TSMFull import *

import GridStateEstimator
reload(GridStateEstimator)
from GridStateEstimator import *

import sonarModel
reload(sonarModel)

## Change this to be your directory
dataDirectory = "/afs/athena.mit.edu/user/l/p/lpk/Desktop/ps13DesignCode/distAngle/"
#dataDirectory = "/Users/lpk/lpkmac/C1/ps13/ps13codeAll/distAngle/"

# Plan is a list of (action, state) pairs
def drawPlan(gw, plan):
    gw.drawWorld(robot.planW, robot.windowWidth)
    lastStep = len(plan) - 1
    for step, (a, s) in enumerate(plan):
        if step == 0:
            color = "DarkGreen"
        elif step == lastStep:
            color = "Gold"
        else:
            color = "Blue"
        gw.drawState(s, robot.planW, robot.windowWidth, color)

## This is a terminating state machine class.  It takes a goal, a
## grid-world, and a planner as argument at initialization time.  The
## state machine plans a path from the current state to the goal, and
## executes a terminating behavior to drive to the first "subgoal"
## location on that path.  When that subgoal is reached, or the driver
## terminates for another reason, the path is re-planned, and then the
## robot drives to the first subgoal on the new path.  When the goal
## location is reached, the behavior terminates.

## IPE stands for "interleaved planning and execution"        
class IPE():

    def __init__(self, goalIndices, gw, planner, locationOnly = False):
        self.goalIndices = goalIndices
        self.gw = gw
        self.planner = planner
        # Do we want to terminate as soon as our location is correct?
        # (Ignore orientation)
        self.locationOnly = locationOnly

    def start(self, input):
        
        #  Update our belief state
        self.gw.updateAndDisplay(input.sonars, input.odometry,
                                 robot.percW,
                                 robot.beliefW,
                                 robot.windowWidth)
        
        #  Assume we're at the most likely pose
        currentPose = self.gw.mostLikelyPose()
        self.currentIndices = self.gw.poseToIndicesAndDir(currentPose)
        
        # Make a plan from there to the goal
        plan = self.planner.plan(initialState = self.currentIndices,
                                 goalTest = lambda x: x == self.goalIndices)
        if plan:
            drawPlan(self.gw, plan)
        else:
            # Couldn't get there:  draw start and goal
            drawPlan(self.gw, [(None, self.currentIndices),
                               (None, self.goalIndices)])
            
        if plan and len(plan) > 1:
            # Return a terminating behavior to take the first action
            # plan[1] is the first real step,  [1] is the location part

            # Subgoal is in planning world coordinates
            subgoal = self.gw.indicesAndDirToPose(plan[1][1])
            # Compute the vector to the goal, in robocentric coordinates
            desiredDeltaXYTh = globalPoseToLocalPose(currentPose, subgoal)
            # Now, compute a subgoal in the odometry frame, for the driver
            odometryGoal = localPoseToGlobalPose(input.odometry,
                                                 desiredDeltaXYTh)
            # Here's where we want to go
            self.macroAction = XYThDriverOA(odometryGoal, robot.goalEpsilon)
        else:
            # Wait 5 steps and see if we've been kidnapped
            self.macroAction = Wait(5)
        # !! We should wander at random if we don't think we can get
        # there from here
        print "Replanned at step", robot.step

        # Start the macroAction
        self.macroAction.start(input)

    def step(self, input):
        replanFrequency = 50
        if self.macroAction.done() or robot.step % replanFrequency == 0:
            self.start(input)
        return self.macroAction.step(input)

    def currentOutput(self):
        return self.macroAction.currentOutput()

    def done(self):
        if self.locationOnly:
            return self.currentIndices[0:2] == self.goalIndices[0:2]
        else:
            return self.currentIndices == self.goalIndices

######################################################################
###
###          Brain methods
###
######################################################################
def setup():
    robot.step = 0
    # Set up IO to just use the robot sensors and wheel motors
    # Use discrete motor outputs so the robot stops if the brain
    # computes for a long time

    motorOutputStepLength = 0.2
    initIO(sonarDistances, pose,
           lambda f, r: discreteMotorOutput(f, r, motorOutputStepLength),
           loadNidaq = False)   # Change to true to use nidaq
    
    # What world we're working on, and where we want to be
#     obstacles = boxWorldBoxes
#     sensorModel = dataDirectory + "boxWorld20.dat"
#     gridN = 20
#     (xmin, xmax, ymin, ymax) = (0.0, 3.05, 0.0, 3.05)
#     goalPose1 = (0.6, 0.6, math.pi / 2)   # world coords
#     goalPose2 = (2.5, 2.5, math.pi)   # world coords

#     obstacles = emptyBoxes
#     sensorModel = dataDirectory + "emptyWorld20.dat"
#     gridN = 20
#     (xmin, xmax, ymin, ymax) = (0.0, 3.05, 0.0, 3.05)
#     goalPose1 = (0.6, 0.6, math.pi / 2)   # world coords
#     goalPose2 = (2.5, 2.5, math.pi)   # world coords

#     obstacles = sheBoxes
#     sensorModel = dataDirectory + "sheWorld20.dat"
#     gridN = 20
#     (xmin, xmax, ymin, ymax) = (0.0, 4.0, 0.0, 4.0)
#     goalPose1 = (0.5, 0.5, math.pi / 2)   # world coords
#     goalPose2 = (3.5, 0.5, math.pi / 2)   # world coords

    obstacles = boxWorldSmallBoxes
    sensorModel = dataDirectory + "boxWorldSmall20.dat"
    gridN = 20
    (xmin, xmax, ymin, ymax) = (0.0, 3.05, 0.0, 3.05)
    goalPose1 = (0.5, 0.5, math.pi / 2)   # world coords
    goalPose2 = (0.5, 2.7, math.pi / 2)   # world coords

    # Basic parameters
    (xrange, yrange) = (xmax-xmin, ymax-ymin)
    (wxmin, wxmax, wymin, wymax) = (xmin - 0.05*xrange, xmax + 0.05*xrange,
                                    ymin - 0.05*yrange, ymax +
                                    0.05*yrange)
    (windowWidth, windowHeight) = (400, 400)
    robot.windowWidth = windowWidth # we need this elsewhere
    # Get within 1/5 of a grid cell of the goal
    robot.goalEpsilon = (xmax - xmin) / (3 * gridN)

    # Make windows
    robot.planW = DrawingWindow(windowWidth, windowHeight,
                                wxmin, wxmax, wymin, wymax, "Plan")
    robot.beliefW = DrawingWindow(windowWidth, windowHeight,
                               wxmin, wxmax, wymin, wymax, "Belief")
    robot.percW = DrawingWindow(windowWidth, windowHeight,
                                wxmin, wxmax, wymin, wymax, "P(O | S)")

    # Make state estimator and planner, do some initial stuff
    gw = GridStateEstimator(Map(obstacles), xmin, xmax, ymin, ymax, gridN,
                            sensorModel, sonarModel.obsProb)
    gwPlanner = Planner(gw)
    
    gw.drawWorld(robot.planW, windowWidth)
    gw.drawBelief(robot.beliefW, windowWidth)
    gw.initPose(pose())   # Remember current odometry reading

    # Interleaved planning and execution to go to the goal 
    robot.planExecute = \
        IPE(gw.poseToIndicesAndDir(goalPose1), gw, gwPlanner,
            locationOnly = True)     # Ignore orientation

    robot.planExecute.start(SensorInput())

def step():
    robot.step += 1
    if robot.planExecute.done():
        print "brain done"
        doAction(Action())
    else: 
        doAction(robot.planExecute.step(SensorInput()))

def shutdown():
    if robot.percW:
        robot.percW.destroy()
    if robot.beliefW:
        robot.beliefW.destroy()
    if robot.planW:
        robot.planW.destroy()


