import math

import utilities
reload(utilities)
from utilities import *

import io
reload(io)
from io import *

import TSMFull
reload (TSMFull)
from TSMFull import *

# Define new terminating behavior that goes to an (x, y, th) pose in
# the frame defined by the odometry.  Only good for local motions Init
# and start should not be delayed from one another, or nothing will
# make sense.

class XYThDriver:
   forwardGain = 2.0
   rotationGain = 2.0
   angleEps = 0.05

   def __init__(self, goalPose, distEps):
      self.distEps = distEps
      self.goalPose = goalPose

   def start(self, input):
      (sonars, pose) = (input.sonars, input.odometry)
      self.currentPose = pose

   def step(self, input):
      self.currentPose = input.odometry
      return self.currentOutput()

   def currentOutput(self):
      (px, py, ptheta) = self.currentPose
      (gx, gy, gtheta) = self.goalPose

      if nearPose2((px,py), (gx,gy), self.distEps):
         if nearAngle(ptheta, gtheta, self.angleEps):
            # We're done
            return Action()
         else:
            # XY is correct; rotate to final heading
            return Action(rvel = fixAnglePlusMinusPi(gtheta - ptheta)*\
                                 self.rotationGain)
      else:
         # Need to translate
         headingTheta = angleToPoint((px,py),(gx,gy))
         headingDeltaTheta = fixAnglePlusMinusPi(headingTheta - ptheta)
         if nearAngle(ptheta, headingTheta, self.angleEps):
            # Move forward
            r = poseDist2((px, py), (gx, gy))
            return Action(fvel = r * self.forwardGain)
         else:
            # Rotate to point toward goal
            return Action(rvel = headingDeltaTheta * self.rotationGain)
   
   def done(self):
      return nearPose(self.currentPose, self.goalPose,
                      self.distEps, self.angleEps)

# Define a new terminating behavior that goes to an (x,y,th) pose
# Terminate if blocked
blockedThreshold = 0.1

class XYThDriverOA (XYThDriver):
    def start(self, input):
        self.blocked = min(input.sonars) < blockedThreshold
        XYThDriver.start(self, input)

    def step(self, input):
        self.blocked = min(input.sonars) < blockedThreshold
        return XYThDriver.step(self, input)

    # Terminate if we're too close to any obstacle
    def done(self):
        return XYThDriver.done(self) or self.blocked

class Stop:
    def __init__(self):
        pass
    def start(self, input = None):
        pass
    def step(self, input = None):
        return self.currentOutput()
    def currentOutput(self):
        return Action()
    def done(self):
        return True

class Wait:
    def __init__(self, waitingTime):
        self.waitingTime = waitingTime
    def start(self, input = None):
        self.currentTime = self.waitingTime
    def step(self, input = None):
        print "waiting"
        self.currentTime = self.currentTime - 1
        return self.currentOutput()
    def currentOutput(self):
        return Action()
    def done(self):
        return self.waitingTime <= 0


# Drive forward a distance d from the pose at initialization time
class ForwardTSM:
    forwardGain = 3.0
    distTargetEpsilon = 0.05

    def __init__(self, delta):
        self.deltaDesired = delta
        
    def done(self):
        return within(poseDist2(self.startingPose, self.currentPose),
                      self.deltaDesired,
                      self.distTargetEpsilon)
    
    def step(self, input):
        (x, y, theta) = input.odometry
        self.currentPose = (x, y)
        return self.currentOutput()
    
    def start(self, input):
        (x, y, theta) = input.odometry
        print "Starting forward", self.deltaDesired
        self.startingPose = (x, y)
        self.currentPose = (x, y)
        
    def currentOutput(self):
        error = self.deltaDesired - \
                     poseDist2(self.startingPose, self.currentPose)
        return Action(fvel = self.forwardGain * error)

class BackwardTSM (ForwardTSM):
    def currentOutput(self):
        error = self.deltaDesired - \
                     poseDist2(self.startingPose, self.currentPose)
        return Action(fvel = -self.forwardGain * error)    

class RotateTSM:
    rotationalGain = 3.0
    angleEpsilon = 0.1
    
    def __init__(self, headingDelta):
        self.headingDelta = headingDelta
        
    def done(self):
        return nearAngle(self.desiredHeading, self.currentHeading,
                         self.angleEpsilon)
    
    def step(self, input):
        (x, y, theta) = input.odometry
        self.currentHeading = theta
        return self.currentOutput()
    
    def start(self, input):
        (x, y, theta) = input.odometry
        print "Starting to rotate", self.headingDelta
        self.currentHeading = theta
        self.desiredHeading = fixAnglePlusMinusPi(theta + self.headingDelta)

    def currentOutput(self):
        return Action(rvel = self.rotationalGain * \
                fixAnglePlusMinusPi(self.desiredHeading - self.currentHeading))

stepLength = 0.1

def dance():
    return RepeatTSM(SequentialTSM([ForwardTSM(stepLength),
                                    BackwardTSM(stepLength),
                                    RotateTSM(math.pi / 2)]), 4)

def chaCha():
    return RepeatTSM(SequentialTSM([ForwardTSM(stepLength),
                                    BackwardTSM(stepLength)]), 4)
