import time

######################################################################
###
###          Input / Output
###
######################################################################
delay = 0.01

leftEyeIndex = 6
rightEyeIndex = 7
neckPotIndex = 5

def readNIDAQ():
    if rp:
        result = rp()[0]
        time.sleep(delay)
        return result
    else:
        raise Exception('NIDAQ not initialized for reading')

def writeNIDAQ(v):
    if wp:
        wp(v)
        time.sleep(delay)
    else:
        raise Exception('NIDAQ not initialized for writing')

def readNIDAQInputs():
    inputs = readNIDAQ()
    neckPot = inputs[neckPotIndex]
    leftEye = inputs[leftEyeIndex]
    rightEye = inputs[rightEyeIndex]
    return (neckPot, leftEye, rightEye)

global sonarDistancesG, poseG, motorOutputG, rp, wp, useNidaq

def initIO(sr, pr, mo, loadNidaq = False):
    global sonarDistancesG, poseG, motorOutputG, rp, wp, useNidaq, useRobot
    sonarDistancesG = sr
    poseG = pr
    motorOutputG = mo
    useNidaq = loadNidaq
    if loadNidaq:
        import nidaq
        rp = nidaq.initNIDAQ()
        wp = nidaq.initNIDAQOutput()

class SensorInput:
    def __init__(self):
        global sonarDistancesG, poseG
        self.sonars = sonarDistancesG()
        self.odometry = poseG()
        if useNidaq:
            self.nidaqInputs = readNIDAQInputs()

referenceVoltage = 5.0
class Action:
    def __init__(self, fvel = 0.0, rvel = 0.0, 
                 voltage = referenceVoltage):
        self.fvel = fvel
        self.rvel = rvel
        self.voltage = voltage

def doAction(a):
    global motorOutputG
    if useNidaq:
        writeNIDAQ(a.voltage)      
    motorOutputG(a.fvel, a.rvel)

